from GRANTA_MIScriptingToolkit import granta as mpy
from ProcessResultsJson import create_report
import getpass
import json


class DatabaseSummarizer:
    """Summarize the contents of a Database and store the results as a json"""
    def __init__(self, database: mpy.Database):
        """
        Initialise class by providing a Database object
        :param database: Database to be summarized
        """
        self.db = database
        self.summary = {}

    def __repr__(self):
        return '<DatabaseSummarizer db: {}; summarized? {}>'.format(self.db.name, bool(self.summary))

    def generate_summary(self):
        """Populate summary property with information, table by table."""
        for tab in self.db.tables:
            print('processing table: {}'.format(tab.name))
            self.summary[tab.name] = self.process_table(tab)

    def print_summary(self):
        """Print the summary to screen"""
        print(self.summary)

    def save_summary_as_json(self, path: str = './', name: str = None):
        """
        Save summary to a json file.
        :param path: path to the save destination
        :param name: name of the file to be saved (if it does not end in .json, .json will be appended
        """
        if name is None:
            name = '{}_results.json'.format(self.db.db_key)
        elif name[-5:] != '.json':
            name = name + '.json'
        with open(path+name, 'w') as f:
            json.dump(self.summary, f)

    def save_summary_as_docx(self, path: str = './', name: str = None):
        """
        Save summary to a docx file.
        :param path: path to the save destination
        :param name: name of the file to be saved (if it does not end in .json, .json will be appended
        """
        assert path is not None
        assert path[-1] == '/'
        if name is None:
            name = '{}_results'.format(self.db.db_key)

        create_report(self.summary, name, path)

    def process_table(self, tab: mpy.Table):
        """
        Create a dictionary for the Table containing information about the records within.
        Table must be one of the tables in the database used to initialise this instance.

        :param tab: mpy.Table object
        :return: dictionary of floats and strings
        """
        attribute_populations = {a: 0 for a in tab.attributes}
        links_populations = {rlg.name: 0 for rlg in self.db.record_link_groups}

        print('    Fetching all records...')
        all_recs = tab.all_records(include_folders=False, include_generics=True)
        num_recs = len(all_recs)
        print('    Fetch complete')

        print('    Fetching all attributes...')
        tab.bulk_fetch(all_recs)
        print('    Fetch complete')

        print('    Fetching all links...')
        tab.bulk_link_fetch(all_recs)
        print('    Fetch complete')

        partial_recs = 0
        full_recs = 0
        empty_recs = 0

        print('    Processing all records...')
        for r in all_recs:
            full, partial, error_in_record = self.calc_population_of_record(attribute_populations, r)
            if not error_in_record:
                if full:
                    full_recs += 1
                    partial = False
                if partial:
                    partial_recs += 1
                if not partial and not full:
                    empty_recs += 1
                for rlg, lv in r.links.items():
                    links_populations[rlg] += len(lv)
        print('    Processing complete')

        return {'No. records': num_recs, 'No. full records': full_recs, 'No. partial records': partial_recs,
                'No. empty records': empty_recs, 'attribute populations': attribute_populations,
                'link populations': links_populations}

    @staticmethod
    def calc_population_of_record(attribute_pops: dict, r: mpy.Record):
        """
        Calculate whether a record is populated and by how much. There are 3 categories: full, partially full,
        and empty.

        :param attribute_pops: dictionary to store results
        :param r: record to be checked
        :return: the bools
        """
        full, partial = True, False
        could_not_check_an_attribute = False
        for a, v in r.attributes.items():
            try:
                empty = v.is_empty()
                if not empty:
                    attribute_pops[a] += 1
                    partial = True
                else:
                    full = False
            except OSError as error:
                could_not_check_an_attribute = True
        return full, partial, could_not_check_an_attribute


if __name__ == '__main__':
    password = getpass.getpass()
    s = mpy.Session('localhost/mi_servicelayer', 'LocAdmin', password)

    db_string = [db.name+'\n' for db in s.dbs]
    db_set = {db.name for db in s.dbs}
    title = 'Database Summarizer\nThe following databases are available, please select one'
    print('-'*20)
    print(title)
    print('-'*20)
    print(''.join(db_string))
    print('-'*20)
    entry = ''
    while entry not in db_set:
        entry = input('Enter a database name: ')
        if entry not in db_set:
            print('Sorry "{}" is not a valid database name, please try again.'.format(entry))

    db = s.get_db(entry)
    dbz = DatabaseSummarizer(db)
    dbz.generate_summary()
    dbz.save_summary_as_docx()
